import datetime
import dataclasses
from typing import Optional, Any, List

from redasql.constants import OperatorType


@dataclasses.dataclass(frozen=True)
class CommandArgs:
    api_key: Optional[str]
    endpoint: Optional[str]
    data_source_name: Optional[str]
    proxy: Optional[str]
    ignore_rc: bool
    debug: bool

    def to_dict(self):
        return dataclasses.asdict(self)


@dataclasses.dataclass(frozen=True)
class DataSourceResponse:
    """
    response for datasources api
    {
        'name': 'metadata',
        'pause_reason': None,
        'syntax': 'sql',
        'paused': 0,
        'view_only': False,
        'type': 'pg',
        'id': 2
    }
    """
    id: int
    name: str
    syntax: str
    paused: int
    view_only: bool
    type: str
    pause_reason: Optional[str]
    supports_auto_limit: Optional[bool] = None  # over v9 attributes

    @property
    def is_query_results(self):
        return self.type in ('results', 'resultswithparams')

    @classmethod
    def from_response(cls, response: dict):
        return cls(**response)


@dataclasses.dataclass(frozen=True)
class QueryResultColumn:
    friendly_name: str
    type: str
    name: str


@dataclasses.dataclass(frozen=True)
class QueryResultData:
    rows: List[dict]
    columns: List[QueryResultColumn]

    @classmethod
    def from_response(cls, data_response):
        return cls(
            rows=data_response['rows'],
            columns=[
                QueryResultColumn(**c)
                for c in data_response['columns']
            ]
        )


@dataclasses.dataclass(frozen=True)
class QueryResultResponse:
    """
    response for query_result api
    {
        'retrieved_at': '2021-10-22T16:25:33.186Z',
        'query_hash': 'd7d1d45bcb946547b382bb0e7853c185',
        'query': 'select 1;',
        'runtime': 0.00282192230224609,
        'data': {
            'rows': [
                {'1': 1}
            ],
            'columns': [
                {'friendly_name': '1', 'type': 'integer', 'name': '1'}
            ]
        },
        'id': 33,
        'data_source_id': 1
    }
    """
    id: int
    data_source_id: int
    retrieved_at: datetime
    query_hash: str
    query: str
    runtime: float
    data: QueryResultData

    @property
    def rows_count(self):
        return len(self.data.rows)

    @property
    def rows_count_for_display(self):
        if self.rows_count > 1:
            return f'{self.rows_count} rows returned.'
        return f'{self.rows_count} row returned.'

    @property
    def runtime_for_display(self):
        return f'Time: {round(self.runtime, 4)}s'

    @classmethod
    def from_response(cls, response):
        return cls(
            id=response['id'],
            data_source_id=response['data_source_id'],
            retrieved_at=datetime.datetime.fromisoformat(
                response['retrieved_at'].replace('Z', '+00:00')
            ),
            query_hash=response['query_hash'],
            query=response['query'],
            runtime=response['runtime'],
            data=QueryResultData.from_response(response['data']),
        )


@dataclasses.dataclass(frozen=True)
class SchemaResponse:
    """
    response for query_result api
    {
        'schema': [
            {
                'name': 'city',
                'columns': ['ID', 'Name', 'CountryCode', 'District', 'Population']
            },
        ]
    }
    """
    name: str
    columns: List[str]

    @classmethod
    def from_response(cls, response):
        return cls(
            name=response['name'],
            columns=response['columns'],
        )


@dataclasses.dataclass(frozen=True)
class QueryParameterResponse:
    name: str
    title: str
    global_: bool
    value: Any
    type: str
    parent_query_id: int

    @classmethod
    def from_response(cls, response: dict):
        return cls(
            name=response['name'],
            title=response['title'],
            global_=response['global'],
            value=response['value'],
            type=response['type'],
            parent_query_id=response['parentQueryId']
        )



@dataclasses.dataclass(frozen=True)
class QueryResponse:
    query: str
    data_source_id: int
    parameters: List[QueryParameterResponse]

    @classmethod
    def from_response(cls, response):
        return cls(
            query=response['query'],
            data_source_id=response['data_source_id'],
            parameters=[
                QueryParameterResponse.from_response(p)
                for p in response['options']['parameters']
            ],
        )


@dataclasses.dataclass(frozen=True)
class NewAttribute:
    attr_name: str
    value: Any
    method_name: Optional[str] = None
    operator: OperatorType = OperatorType.REPLACE


@dataclasses.dataclass(frozen=True)
class MetaCommandReturnList:
    """
    return object for meta command
    """
    new_attrs: List[NewAttribute]

    def apply(self, target: Any):
        for attribute in self.new_attrs:
            if attribute.operator is OperatorType.REPLACE:
                # handle 'hoge.foo' attr name
                attrs = list(reversed(attribute.attr_name.split('.')))
                new_target = target
                while len(attrs) > 1:
                    attr = attrs.pop()
                    new_target = getattr(new_target, attr)
                setattr(new_target, attrs[0], attribute.value)

            elif attribute.operator is OperatorType.CALL:
                attrs = attribute.attr_name.split('.')
                new_target = target
                while len(attrs) > 1:
                    attr = attrs.pop()
                    new_target = getattr(new_target, attr)
                getattr(new_target, attribute.method_name)(attribute.value)
            else:
                print(f'{attribute.operator} is unknown.')



import requests
import time
from typing import Optional, List

from redasql.dto import (
    DataSourceResponse,
    QueryResultResponse,
    SchemaResponse,
    QueryResponse
)
from redasql.exceptions import (
    QueryRuntimeError,
    DataSourceNotFoundError,
    QueryTimeoutError,
)


class ApiClient:

    def __init__(self, redash_url: str, api_key: str, proxy: str = None, debug: bool = False):
        self.redash_url = redash_url.rstrip('/')
        self.session = requests.Session()
        self.session.headers.update({'Authorization': f'Key {api_key}'})
        self.proxy = None
        if proxy:
            self.proxy = proxy
        if self.proxy:
            self.session.proxies.update({'http': self.proxy, 'https': self.proxy})
        if debug:
            import logging
            import http.client as http_client
            requests_log = logging.getLogger("requests.packages.urllib3")
            requests_log.setLevel(logging.DEBUG)
            fmt = "[DEBUG LOGGING][%(asctime)s] %(levelname)s %(name)s :%(message)s"
            logging.basicConfig(level=logging.DEBUG, format=fmt)
            http_client.HTTPConnection.debuglevel = 1

    def get_data_sources(self) -> List[DataSourceResponse]:
        res = self._get('api/data_sources')
        res_json = res.json()
        return [DataSourceResponse.from_response(ds) for ds in res_json]

    def get_data_source_by_name(self, name: str) -> Optional[DataSourceResponse]:
        for ds in self.get_data_sources():
            if ds.name == name:
                return ds
        raise DataSourceNotFoundError(f'data source name [{name}] is not found.')

    def get_data_source_by_id(self, data_source_id: int) -> Optional[DataSourceResponse]:
        for ds in self.get_data_sources():
            if ds.id == data_source_id:
                return ds
        raise DataSourceNotFoundError(f'data source id [{data_source_id}] is not found.')

    def get_schema(self, data_source_id: int) -> List[SchemaResponse]:
        res = self._get(f'api/data_sources/{data_source_id}/schema')
        if 'schema' not in res.json():
            return []
        return [SchemaResponse.from_response(s) for s in res.json()['schema']]

    def get_queries(self):
        res = self._get('api/queries')
        return res.json()

    def get_query_by_id(self, query_id: int) -> QueryResponse:
        res = self._get(f'api/queries/{query_id}')
        return QueryResponse.from_response(res.json())

    def get_version(self):
        res = self._get('api/session')
        res_json = res.json()
        return res_json['client_config']['version']

    def get_query_result(self, query_result_id: int) -> QueryResultResponse:
        res = self._get(f'api/query_results/{query_result_id}')
        return QueryResultResponse.from_response(res.json()['query_result'])

    def execute_query(
            self, query: str, data_source_id: int, max_age: int = -1, timeout=10*60
    ) -> QueryResultResponse:
        res = self._post(
            'api/query_results',
            json={
                'query': f'{query}',
                'data_source_id': data_source_id,
                'max_age': max_age,
            }
        )
        res_json = res.json()
        # has result
        if 'query_result' in res_json:
            return QueryResultResponse.from_response(res_json['query_result'])

        # no result, wait execution.
        job_id = res_json['job']['id']
        return self._wait_job(job_id, timeout)

    def _wait_job(self, job_id: str, timeout: int):
        retry_counter = 0
        while True:
            res = self._get(f'api/jobs/{job_id}')
            res_json = res.json()

            # 3: success
            if res_json.get('job', {}).get('status') == 3:
                query_result_id = res_json['job']['query_result_id']
                return self.get_query_result(query_result_id)

            # 4: error
            if res_json.get('job', {}).get('status') == 4:
                error_msg = res_json['job']['error']
                raise QueryRuntimeError(error_msg)

            retry_counter += 1
            if retry_counter > timeout:
                raise QueryTimeoutError(f'Wait Query Complete. but not completed[{timeout}].')
            time.sleep(0.1)

    def _get(self, path, **kwargs):
        return self._request('GET', path, **kwargs)

    def _post(self, path, **kwargs):
        return self._request('POST', path, **kwargs)

    def _request(self, method, path, **kwargs):
        url = '{}/{}'.format(self.redash_url, path)
        response = self.session.request(method, url, **kwargs)
        response.raise_for_status()
        return response



import csv
from abc import ABC, abstractmethod
from io import StringIO

import tabulate

from redasql.constants import FormatterType
from redasql.dto import QueryResultResponse

tabulate.PRESERVE_WHITESPACE = True


class Formatter(ABC):
    formatter_type = None

    def __init__(self, result: QueryResultResponse, pivoted: bool):
        self.result = result
        self.pivoted = pivoted
        self.query = result.query
        self.column_name_list = [c.name for c in self.result.data.columns]
        self.rows = self.result.data.rows

    def format(self):
        if self.pivoted:
            return self._format_result_to_column_base()
        return self._format_result_to_row_base()

    @abstractmethod
    def _format_result_to_row_base(self):
        raise NotImplementedError()

    @abstractmethod
    def _format_result_to_column_base(self):
        raise NotImplementedError()


class TableFormatter(Formatter):
    formatter_type = FormatterType.TABLE

    def _format_result_to_row_base(self):
        row_for_tables = []
        for row in self.rows:
            row_for_tables.append([row[c] for c in self.column_name_list])
        table = tabulate.tabulate(
            row_for_tables,
            headers=self.column_name_list,
            tablefmt='psql'
        )
        return table

    def _format_result_to_column_base(self):
        result_string = ''
        max_col_name_length = max([len(c) for c in self.column_name_list])
        for i, row in enumerate(self.rows):
            result_string += f'-[RECORD {i + 1}]' + '-' * max_col_name_length + '\n'
            for col in self.column_name_list:
                result_string += f'{col.rjust(max_col_name_length)}| {row[col]}\n'
        return result_string


class MarkdownFormatter(Formatter):
    formatter_type = FormatterType.MARKDOWN

    def _format_result_to_row_base(self):
        row_for_tables = []
        for row in self.rows:
            row_for_tables.append([row[c] for c in self.column_name_list])
        table = tabulate.tabulate(
            row_for_tables,
            headers=self.column_name_list,
            tablefmt='github'
        )
        return table

    def _format_result_to_column_base(self):
        row_for_tables = []
        for row in self.rows:
            for k, v in row.items():
                row_for_tables.append([k, v])

        table = tabulate.tabulate(
            row_for_tables,
            headers=['colum_name', 'value'],
            tablefmt='github'
        )
        return table


class MarkdownWithSQLFormatter(MarkdownFormatter):
    formatter_type = FormatterType.MARKDOWN_WITH_SQL

    def _format_result_to_row_base(self):
        table = super()._format_result_to_row_base()
        result = f"""
```sql
{self.query}
```

{table}"""
        return result


    def _format_result_to_column_base(self):
        table = super()._format_result_to_column_base()
        result = f"""
```sql
{self.query}
```

{table}"""
        return result


class CSVFormatter(MarkdownFormatter):
    formatter_type = FormatterType.CSV

    def _format_result_to_row_base(self):
        with StringIO() as f:
            writer = csv.DictWriter(f, fieldnames=self.column_name_list)
            writer.writeheader()
            for row in self.rows:
                writer.writerow(row)
            return f.getvalue()

    def _format_result_to_column_base(self):
        print('[warn] csv format not suppported pivot')
        return self._format_result_to_row_base()


def formatter_factory(formatter_type: FormatterType):
    formatter = FORMATTERS.get(formatter_type)
    return formatter


FORMATTERS = {
    FormatterType.TABLE: TableFormatter,
    FormatterType.MARKDOWN: MarkdownFormatter,
    FormatterType.MARKDOWN_WITH_SQL: MarkdownWithSQLFormatter,
    FormatterType.CSV: CSVFormatter,
}



import argparse
import os
import itertools
import re
import sys
import pkg_resources
from pyperclip import copy

from textwrap import dedent
from os.path import expanduser, exists
from typing import Optional, List

from prompt_toolkit import prompt
from prompt_toolkit.history import FileHistory

import redasql.utils as utils
from redasql.api_client import ApiClient
from redasql.constants import FormatterType, CompleterType, SQL_KEYWORDS, OutputType
from redasql.dto import CommandArgs, DataSourceResponse, SchemaResponse
from redasql.exceptions import RedasqlException, InsufficientParametersError, NoDataSourceError
from redasql.metacommand_executor import meta_command_factory
from redasql.result_formatter import formatter_factory
from redasql.completer import RedasqlCompleter


VERSION = pkg_resources.get_distribution('redasql').version


class CompleteData:
    def __init__(self):
        self.keywords = SQL_KEYWORDS
        self.schemas: List[SchemaResponse] = []
        self.data_sources = []
        self.formats = FormatterType.values()
        self.outputs = OutputType.values()

    @property
    def column_names(self):
        return list(
            itertools.chain.from_iterable([schema.columns for schema in self.schemas])
        )

    @property
    def schema_names(self):
        return [schema.name for schema in self.schemas]

    def get_completer_words(self):
        words = list(set(
                self.column_names +
                self.schema_names +
                self.keywords +
                self.data_sources +
                self.formats +
                self.outputs
        ))
        return sorted(words, key=lambda x: len(x))

    def get_completer_meta_dict(self):
        meta_dict = {}
        meta_dict.update({c: CompleterType.TABLE.value for c in self.schema_names})
        meta_dict.update({c: CompleterType.COLUMN.value for c in self.column_names})
        meta_dict.update({c: CompleterType.KEYWORD.value for c in self.keywords})
        meta_dict.update({c: CompleterType.DATA_SOURCE.value for c in self.data_sources})
        meta_dict.update({c: CompleterType.FORMAT.value for c in self.formats})
        meta_dict.update({c: CompleterType.OUTPUT.value for c in self.outputs})
        return meta_dict


class MainCommand:

    def __init__(
        self,
        endpoint: str,
        api_key: str,
        proxy: str,
        data_source_name: Optional[str],
        ignore_rc: bool,
        debug: bool,
    ):
        self.endpoint = endpoint if endpoint else os.environ.get('REDASQL_REDASH_ENDPOINT')
        self.api_key = api_key if api_key else os.environ.get('REDASQL_REDASH_APIKEY')
        if self.endpoint is None or self.api_key is None:
            raise InsufficientParametersError(dedent("""
            "endpoint" and "api key" are absolutely necessary. use args or environment
            """))

        self.proxy = proxy if proxy else os.environ.get('REDASQL_HTTP_PROXY')
        self.ignore_rc = ignore_rc
        self.debug = debug
        self.client = ApiClient(
            redash_url=self.endpoint,
            api_key=self.api_key,
            proxy=self.proxy,
            debug=debug,
        )
        self.pivoted = False
        self.output = OutputType.STDOUT
        self.formatter = formatter_factory(FormatterType.TABLE)
        self.input_buffer = []
        self.complete_data = CompleteData()
        self.history = FileHistory(f'{expanduser("~")}/.redasql.hist')
        self.data_source: Optional[DataSourceResponse] = None

        data_source_list = self.client.get_data_sources()
        self.complete_data.data_sources = [d.name for d in data_source_list]
        if data_source_name:
            self.execute_meta_command_handler(fr'\c {data_source_name}')

    def load_config_from_rc_file(self):
        if self.ignore_rc:
            return
        rc_file_path = os.environ.get('REDASQL_RCFILE') if os.environ.get('REDASQL_RCFILE') \
            else f'{expanduser("~")}/.redasqlrc'
        if not exists(rc_file_path):
            return
        with open(rc_file_path, 'r') as f:
            print(f'** load rc file from {rc_file_path} **')
            for line in f:
                if line:
                    try:
                        self._input_handler(line)
                    except Exception as e:
                        print(e)


    def splash(self):
        print(dedent(f"""
         ____          _       ____   ___  _     
        |  _ \ ___  __| | __ _/ ___| / _ \| |    
        | |_) / _ \/ _` |/ _` \___ \| | | | |    
        |  _ <  __/ (_| | (_| |___) | |_| | |___ 
        |_| \_\___|\__,_|\__,_|____/ \__\_\_____|
                                                 
            - redash query cli tool -

        SUCCESS CONNECT
        - server version {self.get_version()}
        - client version {VERSION}
        """))

    def get_version(self):
        return self.client.get_version()

    def loop(self):
        while True:
            try:
                self.main()
            except RedasqlException as e:
                if self.debug:
                    import traceback
                    print(traceback.format_exc())
                print(e)
                self.input_buffer = []
            except KeyboardInterrupt:
                print('Cancel')
                self.input_buffer = []
            except EOFError:
                print('Sayonara!')
                sys.exit(0)

    def main(self):
        answer = prompt(
            self._get_prompt(),
            history=self.history,
            completer=self._get_completer(),
        )
        self._input_handler(answer)

    def _input_handler(self, input_text: str):
        if self.debug:
            print('buffer', self.input_buffer)
        # ignore empty line.
        if utils.is_empty(input_text):
            return

        self.input_buffer.append(input_text)
        if utils.is_meta_command(input_text):
            self.input_buffer = []
            self.execute_meta_command_handler(input_text)
            return

        if utils.is_sql_end(input_text):
            query = '\n'.join(self.input_buffer)
            self.input_buffer = []
            self.execute_query_handler(query)
            return

    def execute_query_handler(self, query: str):
        if not self.data_source:
            raise NoDataSourceError('Plz set datasource.(use \\c <data source name>)')

        query_result = self._execute_query(
            query=query,
            data_source_id=self.data_source.id
        )
        if query_result.rows_count == 0:
            print(dedent(f"""
            no rows returned.
            {query_result.runtime_for_display}
            """))
            return
        formatted_string = self.formatter(query_result, self.pivoted).format()
        print(formatted_string)
        print(dedent(f"""
        {query_result.rows_count_for_display}
        {query_result.runtime_for_display}
        """))
        if self.output == OutputType.STDOUT_AND_CLIPBOARD:
            copy(formatted_string)

    def execute_meta_command_handler(self, input_string):
        command, *args = re.split(r'\s+', input_string.strip())
        executor = meta_command_factory(command)
        e = executor(self.client, self.data_source, self.pivoted, self.formatter, self.output)
        meta_command_return_list = e.exec(*args)
        if self.debug:
            print(command, args, meta_command_return_list)
        if meta_command_return_list:
            meta_command_return_list.apply(self)

    def _execute_query(self, query: str, data_source_id: int):
        if self.debug:
            print('query', query)
        return self.client.execute_query(
            query=query,
            data_source_id=data_source_id,
            max_age=0,
        )

    def _get_prompt(self):
        data_source_name = '(No DataSource)'
        if self.data_source:
            data_source_name = f'{self.data_source.name}'
        suffix = '-' if self.input_buffer else '='
        return f'{data_source_name}{suffix}# '

    def _get_completer(self):
        return RedasqlCompleter(
            latest_inputs=self.input_buffer,
            words=self.complete_data.get_completer_words(),
            meta_dict=self.complete_data.get_completer_meta_dict()
        )


def main():
    args = init()
    try:
        command = MainCommand(**args.to_dict())
        command.splash()
        command.load_config_from_rc_file()
    except Exception as e:
        print(f'[ERROR] {e}\n')
        sys.exit(1)

    command.loop()


def init():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        '-k',
        '--api-key',
        help='redash api key',
        default=None,
    )
    parser.add_argument(
        '-s',
        '--server-host',
        help='redash host i.e) https://your-redash-server/',
        default=None,
    )
    parser.add_argument(
        '-d',
        '--data-source',
        help=dedent("""
        initial datasource name.
        if not set, no datasource selected.
        """),
        default=None
    )
    parser.add_argument(
        '-p',
        '--proxy',
        help=dedent("""
        if you need proxy connection, set them
        i.e) http://user:pass@proxy-host:proxy-port'
        """),
        default=None,
    )
    parser.add_argument(
        '--ignore-rc',
        help=dedent("""
        ignore rc file
        """),
        action='store_true',
        default=False,
    )
    parser.add_argument(
        '--debug',
        help=dedent("""
        debug mode on
        """),
        action='store_true',
        default=False,
    )
    args = parser.parse_args()
    return CommandArgs(
        api_key=args.api_key,
        endpoint=args.server_host,
        data_source_name=args.data_source,
        proxy=args.proxy,
        ignore_rc=args.ignore_rc,
        debug=args.debug,
    )


if __name__ == '__main__':
    main()



import re
from typing import List
from prompt_toolkit.completion import FuzzyWordCompleter
from redasql.constants import CompleterType


class RedasqlCompleter(FuzzyWordCompleter):

    def __init__(self, latest_inputs: List[str], *args, **kwargs):
        self.latest_input_string = ' '.join(latest_inputs)
        super().__init__(*args, **kwargs)

    def get_completions(self, document, complete_event):
        targets = CompleterType.normal_completer_types()
        target_text = document.text[:document.cursor_position]
        if self.check_last_word(target_text, 'from'):
            targets = [CompleterType.TABLE.value]

        if self.check_last_word(target_text, 'select'):
            targets = [CompleterType.COLUMN.value]

        if self._is_in_meta(target_text, r'\c'):
            targets = [CompleterType.DATA_SOURCE.value]

        if self._is_in_meta(target_text, r'\f'):
            targets = [CompleterType.FORMAT.value]

        if self._is_in_meta(target_text, r'\d'):
            targets = [CompleterType.TABLE.value]

        if self._is_in_meta(target_text, r'\o'):
            targets = [CompleterType.OUTPUT.value]

        for completer in super().get_completions(document, complete_event):
            if targets and completer.display_meta_text not in targets:
                continue
            yield completer

    def check_last_word(self, text: str, expect_word: str):
        target = self.latest_input_string + ' ' + text
        target.replace('\\n', ' ')
        target_list = re.split(r'\s+', target)
        # last 1 word is current input. so ignore.
        if len(target_list) > 2:
            return target_list[-2].lower().strip().endswith(expect_word)
        return False

    def _is_in_meta(self, text: str, target):
        return text.startswith(target)



def is_empty(target_str: str):
    return target_str.strip() == ''


def is_sql_end(target_str: str):
    return target_str.strip().endswith(';')


def is_meta_command(target_str: str):
    return target_str.strip().startswith('\\')



import sys
import fnmatch
import re

from typing import Optional
from abc import ABC, abstractmethod

from prompt_toolkit import prompt

from redasql.api_client import ApiClient
from redasql.constants import (
    OperatorType,
    FormatterType,
    OutputType,
)
from redasql.dto import MetaCommandReturnList, NewAttribute, DataSourceResponse
from redasql.exceptions import (
    InvalidMetaCommand,
    InvalidSettingError,
    InsufficientParametersError,
    NoDataSourceError
)
from redasql.result_formatter import Formatter, formatter_factory


class MetaCommandBase(ABC):
    def __init__(self,
                 client: ApiClient,
                 data_source: DataSourceResponse,
                 pivoted: bool,
                 formatter: Formatter,
                 output: OutputType):
        self.client = client
        self.data_source = data_source
        self.pivoted = pivoted
        self.formatter = formatter
        self.output = output

    @staticmethod
    @abstractmethod
    def help_text():
        raise NotImplemented()

    @abstractmethod
    def exec(self, *args, **kwargs) -> Optional[MetaCommandReturnList]:
        pass


class HelpCommandExecutor(MetaCommandBase):

    @staticmethod
    def help_text():
        return 'HELP META COMMANDS.'

    def exec(self, *args, **kwargs):
        for command, executor in EXECUTORS.items():
            print(f'{command}: {executor.help_text()}')


class DescribeCommandExecutor(MetaCommandBase):

    @staticmethod
    def help_text():
        return 'DESCRIBE TABLE'

    @staticmethod
    def _is_match(search_schema_name, target_schema_name):
        """
        support wildcard(*)
        """
        return fnmatch.fnmatch(target_schema_name.lower(), search_schema_name.lower())

    def exec(self, schema_name: str = None, *args, **kwargs) -> Optional[MetaCommandReturnList]:
        if not self.data_source:
            raise NoDataSourceError('Plz set datasource.(use \\c <data source name>)')

        if self.data_source.is_query_results:
            return self._exec_for_query_results(schema_name)
        else:
            return self._exec_for_normal(schema_name)

    def _exec_for_normal(self, schema_name: str = None):
        schemas = self.client.get_schema(self.data_source.id)
        # show all tables
        if not schema_name:
            print(f'## schemas in {self.data_source.name}')
            for schema in schemas:
                print(f'- {schema.name}')
            return

        for schema in schemas:
            if self._is_match(schema_name, schema.name):
                print(f'## {schema.name}')
                for col in schema.columns:
                    print(f'- {col}')
        return

    def _exec_for_query_results(self, schema_name: str = None):
        """
        QueryResultsはdescribe非対応なので、空クエリを投げてCOL情報を
        取得する。
        """
        if not schema_name:
            print('query results not support show all schemas.')
        response = self.client.execute_query(
            query=f'select * from {schema_name} where 1 = 2;',
            data_source_id=self.data_source.id
        )
        print(f'## {schema_name}')
        for col in response.data.columns:
            print(f'- {col.name}')
        return


class ConnectCommandExecutor(MetaCommandBase):

    @staticmethod
    def help_text():
        return 'SELECT DATASOURCE.'

    def exec(self, data_source_name: str = None, *args, **kwargs) -> Optional[MetaCommandReturnList]:
        if not data_source_name:
            data_sources = self.client.get_data_sources()
            for ds in data_sources:
                print(f'{ds.name}:{ds.type}')
            complete_sources = NewAttribute(
                attr_name='complete_data.data_sources',
                value=[d.name for d in data_sources],
            )
            return MetaCommandReturnList(
                new_attrs=[complete_sources]
            )
        data_source = self.client.get_data_source_by_name(data_source_name)
        schemas = self.client.get_schema(data_source_id=data_source.id)
        new_data_source = NewAttribute(
            value=data_source,
            attr_name='data_source'
        )
        new_complete_sources = NewAttribute(
            attr_name='complete_data.schemas',
            value=schemas,
        )
        return MetaCommandReturnList(
            new_attrs=[
                new_data_source,
                new_complete_sources,
            ]
        )


class ChangePivotCommandExecutor(MetaCommandBase):

    @staticmethod
    def help_text():
        return 'QUERY RESULT TOGGLE PIVOT.'

    def exec(self, *args, **kwargs) -> MetaCommandReturnList:
        new_pivoted = not self.pivoted
        print(f'set pivoted [{new_pivoted}]')
        return MetaCommandReturnList(
            new_attrs=[NewAttribute(
                value=not self.pivoted,
                attr_name='pivoted'
            )]
        )


class FormatterChangeExecutor(MetaCommandBase):

    @staticmethod
    def help_text():
        return f'CHANGE RESULT FORMATTER {FormatterType.values()}.'

    def exec(self, formatter_type_name: str = None, *args, **kwargs) -> Optional[MetaCommandReturnList]:
        if not formatter_type_name:
            print(f'current formatter is [{self.formatter.formatter_type.value}]')
            return

        if not FormatterType.is_valid_formatter_type_name(formatter_type_name):
            raise InvalidSettingError(
                f'{formatter_type_name} is not valid formatter type.(chose in {FormatterType.values()})'
            )
        new_formatter_type = FormatterType(formatter_type_name)

        print(f'set formatter [{new_formatter_type.value}]')
        return MetaCommandReturnList(
            new_attrs=[NewAttribute(
                value=formatter_factory(new_formatter_type),
                attr_name='formatter'
            )]
        )


class LoadQueryExecutor(MetaCommandBase):

    @staticmethod
    def help_text():
        return 'LOAD QUERY FROM REDASH.'

    def exec(self, query_id: str = None, *args, **kwargs):
        if not query_id:
            raise InsufficientParametersError('query_id is not specifed.')
        if not query_id.isdigit():
            raise InvalidSettingError(f'[{query_id}] is invalid query_id.')
        query_response = self.client.get_query_by_id(int(query_id))
        query = query_response.query
        for parameter in query_response.parameters:
            answer = prompt(f'{parameter.name}? > ')
            query = self._resplace_parameter(query, parameter.name, answer)
        print(query)
        # set buffer for query execution and add history
        new_attrs = [
            NewAttribute(
                value=[query],
                attr_name='input_buffer',
                operator=OperatorType.REPLACE,
            ),
            NewAttribute(
                value=query,
                attr_name='history',
                method_name='append_string',
                operator=OperatorType.CALL,
            ),
        ]
        # change datasource for query execution.
        if query_response.data_source_id != self.data_source.id:
            data_source = self.client.get_data_source_by_id(query_response.data_source_id)
            print(f'[warn] datasource change to [{data_source.name}]')
            new_attrs.append(
                NewAttribute(
                    value=data_source,
                    attr_name='data_source',
                    operator=OperatorType.REPLACE,
                )
            )

        return MetaCommandReturnList(
            new_attrs=new_attrs
        )

    @staticmethod
    def _resplace_parameter(base_str: str, replace_keyword, replace_value):
        return re.sub(f'{{{{ *{replace_keyword} *}}}}', replace_value, base_str)


class QuitCommandExecutor(MetaCommandBase):

    @staticmethod
    def help_text():
        return 'EXIT.'

    def exec(self, *args, **kwargs):
        print('Sayonara!')
        sys.exit(0)


class ChangeOutputTypeCommandExecutor(MetaCommandBase):

    @staticmethod
    def help_text():
        return f'CHANGE THE OUTPUT DESTINATION TO {OutputType.values()}.'

    def exec(self, output_type_name: str = None, *args, **kwargs) -> Optional[MetaCommandReturnList]:
        if not output_type_name:
            print(f'current  is [{self.output}]')
            return

        if not OutputType.is_valid_output_type_name(output_type_name):
            raise InvalidSettingError(
                f'{output_type_name} is not valid output type.(chose in {OutputType.values()})'
            )
        new_output_type = OutputType(output_type_name)

        print(f'set output [{new_output_type.value}]')
        return MetaCommandReturnList(
            new_attrs=[NewAttribute(
                value=new_output_type,
                attr_name='output'
            )]
        )


def meta_command_factory(command: str):

    executor = EXECUTORS.get(command)
    if not executor:
        raise InvalidMetaCommand(f'{command} is not a valid meta command.')

    return executor


EXECUTORS = {
    r'\?': HelpCommandExecutor,
    r'\q': QuitCommandExecutor,
    r'\d': DescribeCommandExecutor,
    r'\c': ConnectCommandExecutor,
    r'\x': ChangePivotCommandExecutor,
    r'\f': FormatterChangeExecutor,
    r'\l': LoadQueryExecutor,
    r'\o': ChangeOutputTypeCommandExecutor,
}



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['redasql']

package_data = \
{'': ['*']}

install_requires = \
['prompt-toolkit>=3.0.20,<4.0.0',
 'pyperclip>=1.8.2,<2.0.0',
 'requests>=2.26.0,<3.0.0',
 'tabulate>=0.8.9,<0.9.0']

entry_points = \
{'console_scripts': ['redasql = redasql.command:main']}

setup_kwargs = {
    'name': 'redasql',
    'version': '2.2.0',
    'description': '',
    'long_description': "# RedaSQL\n![redasql](https://user-images.githubusercontent.com/4572217/138800787-b9525acd-8ab1-4f35-a762-948244b9caee.png)\n\n\nRedaSQL is querying tool for redash.\nI like `psql`(PostgreSQL CLI). so redasql resemble psql in some respects.\n\n## Install\n\n```bash\npip install redasql\n```\n\n## supported readsh version\n\nRedaSQL supports Redash v8, v9 and v10.RedaSQL supports Redash v8, v9 and v10.\n\n## How To Use\n\nredasql need some arguments or environment variables.\nredasql prioritizes arguments over environment variables.\n\n\n| argument         | env                     | mean                                                                                                        |required|\n|------------------|-------------------------|-------------------------------------------------------------------------------------------------------------|---|\n| -k/--api-key     | REDASQL_REDASH_APIKEY   | API KEY(user api key)                                                                                       |True|\n| -s/--server-host | REDASQL_REDASH_ENDPOINT | Redash server hostname. ex) https://your.redash.server.host/                                                |True|\n| -p/--proxy       | REDASQL_HTTP_PROXY      | if your redash server restricted by Proxy, set url format. ex)http://user:pass@your.proxy.server:proxy-port |False|\n| -d/--data-source | None                    | initial connect datasource name.                                                                            | False                                                                                                       |\n| --ignore-rc      | None                    | ignore `.redasqlrc` file                                                                                    | False(default false) |\n| --debug          | None                    | debug mode                                                                                                  | False(default false) |\n\n\nif you want to use redasql with direnv, rename `.envrc.sample` to `.envrc` and set attributes.\n\n### special commands\n\nredasql has management commands.\n\n```\nmetadata=# \\?\n\\?: HELP META COMMANDS.\n\\q: EXIT.\n\\d: DESCRIBE TABLE\n\\c: SELECT DATASOURCE.\n\\x: QUERY RESULT TOGGLE PIVOT.\n\\f: CHANGE RESULT FORMATTER ['table', 'markdown', 'markdown_with_sql', 'csv'].\n\\l: LOAD QUERY FROM REDASH.\n\\o: CHANGE THE OUTPUT DESTINATION TO ['stdout', 'stdout_and_clipboard'].\n```\n\n### execute query\n\nsee below\n\n#### start\n```\n$ redasql\n\n ____          _       ____   ___  _     \n|  _ \\ ___  __| | __ _/ ___| / _ \\| |    \n| |_) / _ \\/ _` |/ _` \\___ \\| | | | |    \n|  _ <  __/ (_| | (_| |___) | |_| | |___ \n|_| \\_\\___|\\__,_|\\__,_|____/ \\__\\_\\_____|\n\n    - redash query cli tool -\n\nSUCCESS CONNECT\n- server version 8.0.0+b32245\n- client version 0.1.0\n\n(No DataSource)=#\n```\n\n#### connect datasource\n\nuse `\\c data_source_name`. if not provide data_source_name, show all available data sources. \n\n```\n(No DataSource)=# \\c metadata\nmetadata=#\n```\n\n#### describe table\n\nuse `\\d table_name`. if not provide table_name, show all table names. if provide table_name with wildcard(\\*), show describe matched tables.\n\n```\nmetadata=# \\d\naccess_permissions\nalembic_version\n:\nqueries\nquery_results\nquery_snippets\nusers\nvisualizations\nwidgets\nmetadata=# \\d queries\n## queries\n- schedule\n- updated_at\n- api_key\n- name\n- id\n- version\n- is_draft\n- query\n- is_archived\n- tags\n- last_modified_by_id\n- org_id\n- options\n- query_hash\n- description\n- latest_query_data_id\n- search_vector\n- data_source_id\n- schedule_failures\n- created_at\n- user_id\nmetadata=# \\d query_*\n## query_results\n- id\n- data\n- org_id\n- query_hash\n- data_source_id\n- runtime\n- query\n- retrieved_at\n## query_snippets\n- updated_at\n- id\n- description\n- created_at\n- user_id\n- trigger\n- snippet\n- org_id\n\n```\n\n#### execute query\n\nenter your SQL and semicolon.\n\n```bash\nmetadata=# select count(*) from queries;\n+-------+\n| count |\n+-------+\n|  3606 |\n+-------+\n\n1 row returned.\nTime: 0.0159s\n\n```\n\n`\\x` pivot result.\n\n\n\n```\nmetadata=# \\x\nset pivoted [True]\n\nmetadata=# select id, user_id from queries limit 3;\n-[RECORD 1]-------\n     id: 543\nuser_id: 40\n-[RECORD 2]-------\n     id: 717\nuser_id: 40\n-[RECORD 3]-------\n     id: 515\nuser_id: 38\n\n\n3 rows returned.\nTime: 0.0281s\n\n```\n\n### formats\n\nredasql support many formats. `\\f <format_name>` and `\\x`\n\n\n\n#### table format(default)\n\n```\nmetadata=# select id, object_id, org_id, created_at from favorites limit 3;\n\n+------+-------------+----------+--------------------------+\n|   id |   object_id |   org_id | created_at               |\n|------+-------------+----------+--------------------------|\n|    2 |         513 |        1 | 2019-05-22T05:30:17.185Z |\n|    3 |         514 |        1 | 2019-05-22T05:30:19.031Z |\n|    4 |         230 |        1 | 2019-05-22T08:17:12.693Z |\n+------+-------------+----------+--------------------------+\n\n3 rows returned.\nTime: 0.0219s\n```\n\n#### table format(pivoted)\n\n```\nmetadata=# select id, object_id, org_id, created_at from favorites limit 3;\n\n-[RECORD 1]----------\n        id| 2\n object_id| 513\n    org_id| 1\ncreated_at| 2019-05-22T05:30:17.185Z\n-[RECORD 2]----------\n        id| 3\n object_id| 514\n    org_id| 1\ncreated_at| 2019-05-22T05:30:19.031Z\n-[RECORD 3]----------\n        id| 4\n object_id| 230\n    org_id| 1\ncreated_at| 2019-05-22T08:17:12.693Z\n\n\n3 rows returned.\nTime: 0.0223s\n\n```\n\n#### markdown\n\n```\nmetadata=# \\f markdown\nset formatter [markdown]\nmetadata=# select id, object_id, org_id, created_at from favorites limit 3;\n\n|   id |   object_id |   org_id | created_at               |\n|------|-------------|----------|--------------------------|\n|    2 |         513 |        1 | 2019-05-22T05:30:17.185Z |\n|    3 |         514 |        1 | 2019-05-22T05:30:19.031Z |\n|    4 |         230 |        1 | 2019-05-22T08:17:12.693Z |\n\n3 rows returned.\nTime: 0.0207s\n\n```\n\n#### markdown(pivoted)\n\n\n```\nmetadata=# select id, object_id, org_id, created_at from favorites limit 3;\n\n| colum_name   | value                    |\n|--------------|--------------------------|\n| created_at   | 2019-05-22T05:30:17.185Z |\n| org_id       | 1                        |\n| id           | 2                        |\n| object_id    | 513                      |\n| created_at   | 2019-05-22T05:30:19.031Z |\n| org_id       | 1                        |\n| id           | 3                        |\n| object_id    | 514                      |\n| created_at   | 2019-05-22T08:17:12.693Z |\n| org_id       | 1                        |\n| id           | 4                        |\n| object_id    | 230                      |\n\n3 rows returned.\nTime: 0.0106s\n```\n\n#### markdown_with_sql\n\n```\n```sql\nselect id, object_id, org_id, created_at from favorites limit 3;\n``` .\n\n|   id |   object_id |   org_id | created_at               |\n|------|-------------|----------|--------------------------|\n|    2 |         513 |        1 | 2019-05-22T05:30:17.185Z |\n|    3 |         514 |        1 | 2019-05-22T05:30:19.031Z |\n|    4 |         230 |        1 | 2019-05-22T08:17:12.693Z |\n\n3 rows returned.\nTime: 0.0253s\n\n\n```\n\n#### csv\n\n```\nmetadata=# \\f csv\nset formatter [csv]\nmetadata=# select id, object_id, org_id, created_at from favorites limit 3;\n\nid,object_id,org_id,created_at\n2,513,1,2019-05-22T05:30:17.185Z\n3,514,1,2019-05-22T05:30:19.031Z\n4,230,1,2019-05-22T08:17:12.693Z\n\n```\n\n### quit\n\n`ctrl + D` or `\\q` quit redasql.\n\n```\nmetadata=# \\q                                                                                                                                                                        \nSayonara!\n```\n\n### .redasqlrc\n\nif you create `~/.redasqlrc`, redasql reads the file at startup and performs initial settings.\n\n\n## Contribution\n\n### run test\n\n#### start up containers (redash, MySQL, postgresql)\n\nunittest necessary redash test server. use docker-compose.yml in `./tests/docker/v{REDAHS_VERSION}`\n\n - 8.0.2.b37747\n - 9.0.0-beta.b49509\n - 10.0.0.b50363\n\n```bash\n$ cd ./tests/docker/v8.0.2.b37747\n$ docker-compose up -d\n```\n\n#### run test\n\n```bash\n$ python -m unittest discover -s tests/\n```\n",
    'author': 'denzow',
    'author_email': 'denzow@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/denzow/redasql',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



class RedasqlException(Exception):
    """ base exception """


class QueryRuntimeError(RedasqlException):
    """ redash said, sql error """


class QueryTimeoutError(RedasqlException):
    """ too long """


class DataSourceNotFoundError(RedasqlException):
    """ not found datasource name """


class InvalidMetaCommand(RedasqlException):
    """ invalid meta command"""


class InvalidSettingError(RedasqlException):
    """ invalid setting requested. """


class InsufficientParametersError(RedasqlException):
    """ api key and endpoint must be set """


class NoDataSourceError(RedasqlException):
    """ not set datasource """
